#!/usr/bin/perl

use Win32::GUI;
use Win32::GUI::DIBitmap;
use IO::Socket;
use IO::Select;
use Time::HiRes;
use Digest::SHA;

#---- API Use's
use Win32::API;
my $powerzoom          = Win32::API->new('./bin/powerzoom/powerzoom.dll','powerzoom','PPIIIIIIIIIIII', 'I');
my $rep_movsb_repeated = Win32::API->new('./bin/move_dll/move.dll', 'rep_movsb_repeated', 'IIIIII','I');
my $GetGUIThreadInfo   = Win32::API->new('User32.dll', 'GetGUIThreadInfo', 'IP','I');


#==== Global Messageboxes
sub infomessbox_w32($)
{
 Win32::GUI::MessageBox($ARGV[0],$_[0],'VDM: Client Loader',0x40);
 return 1;
}
sub quesmessbox_w32($)
{
 my $answer = Win32::GUI::MessageBox($ARGV[0],$_[0],'VDM: Client Loader',0x24);
 return ($answer == 6)? 1: 0;
}
#sub retrymessbox($)
#{
# my $answer = $mw->messageBox(-title   => 'Virtual Display Manager',
#                              -message => $_[0],
#                              -type    => 'RetryCancel',
#                              -icon    => 'question',
#                              -default => 'Retry');
# return ($answer eq 'Cancel')? 0: 1;
#}

####################################################################
################################################## new socksystem ##
#---- vars
my %sockhash;
my $stream_terminator = "~#END#";
#---- getsock
sub getsock($$)
{
 my $ip   = $_[0];
 my $port = $_[1];
 my $dsock = IO::Socket::INET->new(#Listen    => SOMAXCONN,
                                   PeerAddr  => $ip,
                                   PeerPort  => $port,
                                   #LocalAddr => inet_ntoa(INADDR_ANY),
                                   #LocalPort => $portid,
                                   #Broadcast => 1,
                                   #Type => SOCK_STREAM,
                                   #Blocking => 1,
                                   Proto     => 'tcp');
 return $dsock;
}
#---- getsock + login
sub getsock_andlogin($$$$)
{
 my $ip         = $_[0];
 my $port       = $_[1];
 my $login_name = $_[2];
 my $login_pass = $_[3];

 my $dsock;
 while(not($dsock = &getsock($ip,$port)))
 { return if (! &quesmessbox_w32("Cant connect to Server.\nError message: $!\n\nTry to reconnect ?")); }

 my $publickey;
 if (not(($publickey) = &tcp_slurp($dsock,'LOGINGETPUBLIC')))
 { &infomessbox_w32("Server tcp_slurp problem on login. (1)"); return; }

 my $complexpass = $login_pass.$publickey;
 $complexpass = uc(Digest::SHA::sha1_hex($complexpass)) for (1 .. 7);

 my $pass_answer;
 if (not(($pass_answer) = &tcp_slurp($dsock,'LOGIN~'.$login_name.'/'.$complexpass)))
 { &infomessbox_w32("Server tcp_slurp problem on login. (2)"); return; }

 if ($pass_answer ne 'LOGIN_OK')
 {
  $dsock->close;
  &infomessbox_w32("Authentication problem.\nServer response: ".$pass_answer);
  return;
 }
 return $dsock;
}
#---- tcp_slurp
sub tcp_slurp($$)
{
 my $sock    = $_[0];
 my $tosend  = $_[1];

 my $theselect = IO::Select->new($sock);
 if($theselect->can_write(7))
 {
  #print "can_write\n";
  syswrite($sock,$tosend.$stream_terminator);
  my $sammelrequest;
  while($theselect->can_read(7))
  {
   #print "can_read\n";
   my $data;
   return unless (sysread($sock,$data,1000000));
   $sammelrequest .= $data;
   if ($sammelrequest =~ s/$stream_terminator$//)
   {
    return $sammelrequest;
   }
   #print "must-reread\n";
  }
 }
 close $sock;
 print "TIMEOUT: end_tcp_slurp\n";
 return;
}
#---- login_slurp
sub login_slurp($$$$$)
{
 my $ip          = $_[0];
 my $port        = $_[1];
 my $login_name  = $_[2];
 my $login_pass  = $_[3];
 my $send_string = $_[4];
 my $ip_port     = join(':',$ip,$port);

 if (not defined $sockhash{$ip_port})
 {
   my $server_answer;
   do
   {
    return unless (($sockhash{$ip_port}) = &getsock_andlogin($ip,$port,$login_name,$login_pass));
    if (not(($server_answer) = &tcp_slurp($sockhash{$ip_port},$send_string)))
    { return if (! &quesmessbox_w32("Server tcp_slurp problem.\nTry to reconnect ?")); }
   }
   while (not defined $server_answer);
   return $server_answer;
 }
 else
 {
  if (my ($return_string) = &tcp_slurp($sockhash{$ip_port},$send_string))
  {
   return $return_string;
  }
  else
  {
   return unless (&quesmessbox_w32("Connection to Server broken.\nError message: $!\n\nTry to reconnect ?"));
   my $server_answer;
   do
   {
    return unless (($sockhash{$ip_port}) = &getsock_andlogin($ip,$port,$login_name,$login_pass));
    if (not(($server_answer) = &tcp_slurp($sockhash{$ip_port},$send_string)))
    { return if (! &quesmessbox_w32("Server tcp_slurp problem.\nTry to reconnect ?")); }
   }
   while (not defined $server_answer);
   return $server_answer;
  }
 }
}

#==== filter_oplist
my %filter_optlist = (
"24bit 16M Colors (LAN)"     => [24,"no-filter"],
"16bit 32K Colors"           => [16,"no-filter"],
" 8bit 256 Colors"           => [ 8,"no-filter"],
" 8bit Color Optimized"      => [24,"ColorQuantize",FIQ_WUQUANT,FIF_BMP,0],
" 8bit Color Opt. & Packed"  => [24,"ColorQuantize",FIQ_WUQUANT,FIF_BMP,BMP_SAVE_RLE],
" 8bit Greyscale"            => [24,"ConvertTo8Bits",undef,FIF_BMP,0],
" 8bit Greyscale Packed"     => [24,"ConvertTo8Bits",undef,FIF_BMP,BMP_SAVE_RLE],
" 4bit 16 Colors"            => [ 4,"no-filter"],
" 4bit Greyscale"            => [24,"ConvertTo4Bits",undef,FIF_BMP,0],
" 1bit Threshold 20%"        => [24,"Threshold",chr(49),FIF_BMP,0],
" 1bit Threshold 40%"        => [24,"Threshold",chr(99),FIF_BMP,0],
" 1bit Threshold 60%"        => [24,"Threshold",chr(149),FIF_BMP,0],
" 1bit Threshold 80%"        => [24,"Threshold",chr(199),FIF_BMP,0],
" 1bit Dither Fs"            => [24,"Dither",FID_FS,FIF_BMP,0],
" 1bit Dither Bayer4x4"      => [24,"Dither",FID_BAYER4x4,FIF_BMP,0],
" 1bit Dither Bayer8x8"      => [24,"Dither",FID_BAYER8x8,FIF_BMP,0],
" 1bit Dither Cluster6x6"    => [24,"Dither",FID_CLUSTER6x6,FIF_BMP,0],
" 1bit Dither Cluster8x8"    => [24,"Dither",FID_CLUSTER8x8,FIF_BMP,0],
" 1bit Dither Cluster16x16"  => [24,"Dither",FID_CLUSTER16x16,FIF_BMP,0],
"Jpeg Color q100"           => [24,"ConvertItNot",undef,FIF_JPEG,100],
"Jpeg Color q099"           => [24,"ConvertItNot",undef,FIF_JPEG,99],
"Jpeg Color q098"           => [24,"ConvertItNot",undef,FIF_JPEG,98],
"Jpeg Color q097"           => [24,"ConvertItNot",undef,FIF_JPEG,97],
"Jpeg Color q096"           => [24,"ConvertItNot",undef,FIF_JPEG,96],
"Jpeg Color q094"           => [24,"ConvertItNot",undef,FIF_JPEG,94],
"Jpeg Color q092"           => [24,"ConvertItNot",undef,FIF_JPEG,92],
"Jpeg Color q090"           => [24,"ConvertItNot",undef,FIF_JPEG,90],
"Jpeg Color q080"           => [24,"ConvertItNot",undef,FIF_JPEG,80],
"Jpeg Color q060"           => [24,"ConvertItNot",undef,FIF_JPEG,60],
"Jpeg Color q040"           => [24,"ConvertItNot",undef,FIF_JPEG,40],
"Jpeg Color q020"           => [24,"ConvertItNot",undef,FIF_JPEG,20],
"Jpeg Color q010"           => [24,"ConvertItNot",undef,FIF_JPEG,10],
"Jpeg Color q005"           => [24,"ConvertItNot",undef,FIF_JPEG,5],
"Jpeg Color q001"           => [24,"ConvertItNot",undef,FIF_JPEG,1],
"Jpeg Greyscale q100"       => [24,"ConvertTo8Bits",undef,FIF_JPEG,100],
"Jpeg Greyscale q099"       => [24,"ConvertTo8Bits",undef,FIF_JPEG,99],
"Jpeg Greyscale q098"       => [24,"ConvertTo8Bits",undef,FIF_JPEG,98],
"Jpeg Greyscale q097"       => [24,"ConvertTo8Bits",undef,FIF_JPEG,97],
"Jpeg Greyscale q096"       => [24,"ConvertTo8Bits",undef,FIF_JPEG,96],
"Jpeg Greyscale q094"       => [24,"ConvertTo8Bits",undef,FIF_JPEG,94],
"Jpeg Greyscale q092"       => [24,"ConvertTo8Bits",undef,FIF_JPEG,92],
"Jpeg Greyscale q090"       => [24,"ConvertTo8Bits",undef,FIF_JPEG,90],
"Jpeg Greyscale q080"       => [24,"ConvertTo8Bits",undef,FIF_JPEG,80],
"Jpeg Greyscale q060"       => [24,"ConvertTo8Bits",undef,FIF_JPEG,60],
"Jpeg Greyscale q040"       => [24,"ConvertTo8Bits",undef,FIF_JPEG,40],
"Jpeg Greyscale q020"       => [24,"ConvertTo8Bits",undef,FIF_JPEG,20],
"Jpeg Greyscale q010"       => [24,"ConvertTo8Bits",undef,FIF_JPEG,10],
"Jpeg Greyscale q005"       => [24,"ConvertTo8Bits",undef,FIF_JPEG,5],
"Jpeg Greyscale q001"       => [24,"ConvertTo8Bits",undef,FIF_JPEG,1],
);

#####################################################################
########################################################## UPDATER ##
my $time = Time::HiRes::time;
sub timetest {my $newtime = Time::HiRes::time; print "timediff_loader: ".int(($newtime-$time)*1000)." (".shift().")\n"; $time = $newtime;}
sub windowupdater(@)
{
 my ($client_hwnd,$ip,$port,$login_name,$login_pass,$l,$t,$r,$b,$reload_delay,$jpg_quality) = @_;
 my ($width,$height) = ($r-$l,$b-$t);
 my ($bpp,$filter_function,$filter_flag,$filter_imgformat,$filter_jpgquality) = @{$filter_optlist{$jpg_quality}};
 #$jpg_quality = 70;

 delete $sockhash{$ip.':'.$port};
 my $bmpcache;
 my $scalewidth_old  = 0;
 my $scaleheight_old = 0;
 my $GUITHREADINFO = pack('I',12*4).("    " x (12*4));
 my $forcefullscreen = 0;
 while(my $DC = Win32::GUI::DC::GetDC($client_hwnd))
 {
  #timetest("start");
  my $delay_reload = 2;
  #if(substr(sprintf("%08x",Win32::GUI::GetWindowLong($client_hwnd,-16)),0,1) == 1)#is_visible
  unless(Win32::GUI::IsIconic($client_hwnd))#is_minimized
  {
   #---- ermitteln des zoomfactor und ratio-resize
   my ($scalewidth,$scaleheight) = (Win32::GUI::ScaleWidth($client_hwnd),Win32::GUI::ScaleHeight($client_hwnd));
   my $zoomfactor = $scaleheight / $height;
   if ($scalewidth != $scalewidth_old || $scaleheight != $scaleheight_old)
   {
    $forcefullscreen = 1;
    $GetGUIThreadInfo->Call(0,$GUITHREADINFO);
    unless (unpack("I",substr($GUITHREADINFO,4,4)) & 2)
    {
     $zoomfactor = $scalewidth / $width if ($scaleheight == $scaleheight_old);
     Win32::GUI::Resize($client_hwnd,int($width*$zoomfactor)+(Win32::GUI::Width($client_hwnd)-Win32::GUI::ScaleWidth($client_hwnd)),int($height*$zoomfactor)+(Win32::GUI::Height($client_hwnd)-Win32::GUI::ScaleHeight($client_hwnd)));
     ($scalewidth_old,$scaleheight_old) = (int($width*$zoomfactor),int($height*$zoomfactor));
    }
   }
   #---- SHOOTDISPLAY
   if (my ($dataslurp) = login_slurp($ip,$port,$login_name,$login_pass,"SHOOTDISPLAY~".$l."~".$t."~".$width."~".$height."~".$jpg_quality."~".$forcefullscreen))
   {
    if ($dataslurp =~ m/^([0123])~(\d+)~(\d+)~(.*)/s)
    {
     #print "SHOOTED:$1~$2~$3~length(".length($4).")\n";
     #---- Make BMP
     #my $bmpdata = Win32::GUI::DIBitmap->newFromData($4)->SaveToData(FIF_BMP) if ($1 != 1);
     if ($1 != 1) # if not 'not changed'
     {
      $delay_reload = 1;
      my $dib;
      if (($dib = Win32::GUI::DIBitmap->_newFromData($filter_imgformat,$4,0)))# &&
#          ($dib = $dib->Rescale(int($dib->Width*$zoomfactor),int($dib->Height*$zoomfactor),FILTER_BOX)))
      {
       #$dib = $dib->Rescale(int($dib->Width*$zoomfactor),int($dib->Height*$zoomfactor));
       #my $bmpdata = $dib->ConvertTo4Bits->_saveToData(FIF_BMP,0);
       #my $bmpdata = $dib->ConvertTo16Bits565->_saveToData(FIF_BMP,0); # convert really needed ?
       #my $bmpdata = $dib->Threshold(200)->_saveToData(FIF_BMP,0); # convert really needed ?
       #my $bmpdata = $dib->ColorQuantize()->_saveToData(FIF_BMP,0); # convert really needed ?
       #my $bmpdata = $dib->Dither(FID_CLUSTER8x8)->_saveToData(FIF_BMP,0); # convert really needed ?
       my $bmpdata = $dib->ConvertTo24Bits->_saveToData(FIF_BMP,0);
       undef $dib;
#       $zoomfactor = 1;

       if ($1 == 2)#partially changed
       {
        #---------------------------- bmpfusion
        #my ($bmpcache_x,$bmpcache_y) = unpack('I2',substr($bmpcache,18,8));
        my ($bmpdata_x,$bmpdata_y) = unpack('I2',substr($bmpdata,18,8));
        my ($bmpcache_x,$bmpcache_y) = unpack('I2',substr($bmpcache,18,8));
        #my $bit_per_pixel = unpack('S1',substr($bmpdata,28,2));
        #my $bmpdata_x_mal3 = $bmpdata_x*($bit_per_pixel/8);
        #my $bmpcache_datenbreitex_plusbugbit = int(($bmpcache_x*$bit_per_pixel+31)/32)*4;
        #my $bmpdata_datenbreitex_plusbugbit = int(($bmpdata_x*$bit_per_pixel+31)/32)*4;

        #---- bmpdata_size und in $bmpdata buffer speichern
        my $bmpdata_size_ohneheader = int(($bmpdata_x*24+31) / 32) * 4 * $bmpdata_y;
        substr($bmpdata,14+20,4,pack("L1",$bmpdata_size_ohneheader)); #img-size

        #my $bmpcache_ptr = unpack('I',pack('P',$bmpcache));
        #my $bmpdata_ptr = unpack('I',pack('P',$bmpdata));
        ##my $bmpcache_startptr= $bmpcache_ptr+unpack('I1',substr($bmpcache,10,4))+$bmpcache_datenbreitex_plusbugbit*($bmpcache_y-$bmpdata_y-int($3*$zoomfactor))+int($2*$zoomfactor)*($bit_per_pixel/8);
        #my $bmpcache_startptr= $bmpcache_ptr+unpack('I1',substr($bmpcache,10,4))+$bmpcache_datenbreitex_plusbugbit*($bmpcache_y-$bmpdata_y-$3)+$2*($bit_per_pixel/8);
        #my $bmpdata_startptr = $bmpdata_ptr+unpack('I1',substr($bmpdata,10,4));
        #$rep_movsb_repeated->Call($bmpdata_startptr,$bmpcache_startptr,$bmpdata_x_mal3,$bmpdata_datenbreitex_plusbugbit,$bmpcache_datenbreitex_plusbugbit,$bmpdata_y);
        #$powerzoom->Call($bmpdata,$bmpcache,int($2*$zoomfactor),int($3*$zoomfactor),int(($2+$bmpdata_x)*$zoomfactor),int(($3+$bmpdata_y)*$zoomfactor));
#        $powerzoom->Call($bmpdata,$bmpcache,int($2*$zoomfactor),int($3*$zoomfactor),int(($2+$bmpdata_x)*$zoomfactor),int(($3+$bmpdata_y)*$zoomfactor),$width,$height,$bmpcache_x,$bmpcache_y);
        #print "$3 -> ".($bmpcache_y-int(int(($height-($3+$bmpdata_y))*$bmpcache_y)/$height)-int(int(int($3+$bmpdata_y)*$bmpcache_y)/$height))."\n";
#        print join(" + ",int(int($2*$bmpcache_x)/$width),($bmpcache_y-(int(int(($height-($3+$bmpdata_y))*$bmpcache_y)/$height)) - int(int(int($bmpdata_y)*$bmpcache_y)/$height)),int(int(int($2+$bmpdata_x)*$bmpcache_x)/$width),($bmpcache_y-(int(int(($height-($3+$bmpdata_y))*$bmpcache_y)/$height))),$width,$height,$bmpcache_x,$bmpcache_y,$2,$3)."\n";
#        print join(" + ",int($2*$zoomfactor),int($3*$zoomfactor),int(($2+$bmpdata_x)*$zoomfactor),int(($3+$bmpdata_y)*$zoomfactor),$width,$height,$bmpcache_x,$bmpcache_y)."\n";
        $powerzoom->Call($bmpdata,$bmpcache,int(int($2*$bmpcache_x)/$width),($bmpcache_y-(int(int(($height-($3+$bmpdata_y))*$bmpcache_y)/$height)) - int(int(int($bmpdata_y)*$bmpcache_y)/$height)),int(int(int($2+$bmpdata_x)*$bmpcache_x)/$width),($bmpcache_y-(int(int(($height-($3+$bmpdata_y))*$bmpcache_y)/$height))),$width,$height,$bmpcache_x,$bmpcache_y,$2,$3,$2+$bmpdata_x,$3+$bmpdata_y);
        #$powerzoom->Call($bmpdata,$bmpcache,int($2*$zoomfactor),int($3*$zoomfactor),int(($2+$bmpdata_x)*$zoomfactor),int(($3+$bmpdata_y)*$zoomfactor),$width,$height,$bmpcache_x,$bmpcache_y,$2,$3,$2+$bmpdata_x,$3+$bmpdata_y);
       }
       else
       {
        my ($bmpdata_x,$bmpdata_y) = unpack('I2',substr($bmpdata,18,8));
        my ($bmpcache_x,$bmpcache_y) = (int($bmpdata_x*$zoomfactor),int($bmpdata_y*$zoomfactor));
        my $bmpcache_size = int(($bmpcache_x*24+31) / 32) * 4 * $bmpcache_y + 54 + ($bmpcache_x * 3 * 2);
        vec($bmpcache,$bmpcache_size-1,8) = 0 if (length($bmpcache) < $bmpcache_size);
        #---- bmpdata_size und in $bmpdata buffer speichern
        my $bmpdata_size_ohneheader = int(($bmpdata_x*24+31) / 32) * 4 * $bmpdata_y;
        substr($bmpdata,14+20,4,pack("L1",$bmpdata_size_ohneheader)); #img-size
        #$bmpcache = $bmpdata;# if ($1 != 1);#when not 'not changed'
        #$powerzoom->Call($bmpdata,$bmpcache,-1,-1,$bmpcache_x,$bmpcache_y);
        $xx = $powerzoom->Call($bmpdata,$bmpcache,-1,-1,$bmpcache_x,$bmpcache_y,$bmpdata_x,$bmpdata_y,$bmpcache_x,$bmpcache_y,0,0,$bmpdata_x,$bmpdata_y);
        #print join(",",unpack('I'.(10*4),unpack("P".(10*4),pack("I",$xx))));

        $forcefullscreen = 0;
       }
      }
      #---- zum schliessen bei datenfehler (mit retrypopup) .. achtung tritt ab und zu auf trotz intakter bilddaten .. to solve
      #else { unless (&quesmessbox_w32("Loader Process Failure.\n(Image data is corrupted)\n\nRetry to load ?")) { Win32::GUI::PostMessage($client_hwnd,0x10,0,0); last; } }
     }

     #---- display image
     my $dib = Win32::GUI::DIBitmap->_newFromData(FIF_BMP,$bmpcache,0);
     #$dib = $dib->Rescale(Win32::GUI::ScaleWidth($client_hwnd),Win32::GUI::ScaleHeight($client_hwnd));
     $dib->CopyToDC($DC) if ($dib && Win32::GUI::IsWindow($client_hwnd));
     undef $dib;
    }
    #---- zum warnen wenn DisplayDC nicht an die daten kommt am server
    elsif ($dataslurp eq "-1") { unless (&quesmessbox_w32("Server problem.\n(Server says he cannot open a Display Context Object)\n\nRetry to load ?")) { Win32::GUI::PostMessage($client_hwnd,0x10,0,0); last; } }
    #---- zum schliessen bei datenfehler (mit retrypopup)
    else { unless (&quesmessbox_w32("Loader Process Failure.\n(Data Failure ?)\n\nRetry to load ?")) { Win32::GUI::PostMessage($client_hwnd,0x10,0,0); last; } }
   #---- zum schliessen bei verbindungsfehler (kein warn-popup / kam eh bei slurp)
   } else { Win32::GUI::PostMessage($client_hwnd,0x10,0,0); last; }
  }
  my $released = Win32::GUI::DC::ReleaseDC($client_hwnd,$DC);
  Time::HiRes::sleep($reload_delay*$delay_reload);
  #last;
 }
 undef $bmpcache;
 close $sockhash{$ip.':'.$port};
 #undef $rawcache;
 #threads->exit();
 #threads->detach();
 print "window_killed\n";
}

windowupdater(@ARGV);
#print join(',',@ARGV)."\n";



