#!/usr/bin/perl

use Tk;
use Tk::Icon;
use Tk::NoteBook;
use Tk::Pane;
use Win32::GUI;
use IO::Socket;
use IO::Select;
use Digest::SHA;

#---- glob's
my ($ip,$port,$login_name,$login_pass,$hwnd) = @ARGV;

#==== Global Messageboxes
sub infomessbox_w32($)
{
 Win32::GUI::MessageBox(0,$_[0],'VDM: Window Styles',0x40);
 return 1;
}
sub quesmessbox_w32($)
{
 my $answer = Win32::GUI::MessageBox(0,$_[0],'VDM: Window Styles',0x24);
 return ($answer == 6)? 1: 0;
}

####################################################################
################################################## new socksystem ##
#---- vars
my %sockhash;
my $stream_terminator = "~#END#";
#---- getsock
sub getsock($$)
{
 my $ip   = $_[0];
 my $port = $_[1];
 my $dsock = IO::Socket::INET->new(#Listen    => SOMAXCONN,
                                   PeerAddr  => $ip,
                                   PeerPort  => $port,
                                   #LocalAddr => inet_ntoa(INADDR_ANY),
                                   #LocalPort => $portid,
                                   #Broadcast => 1,
                                   #Type => SOCK_STREAM,
                                   #Blocking => 1,
                                   Proto     => 'tcp');
 return $dsock;
}
#---- getsock + login
sub getsock_andlogin($$$$)
{
 my $ip         = $_[0];
 my $port       = $_[1];
 my $login_name = $_[2];
 my $login_pass = $_[3];

 my $dsock;
 while(not($dsock = &getsock($ip,$port)))
 { return if (! &quesmessbox_w32("Cant connect to Server.\nError message: $!\n\nTry to reconnect ?")); }

 my $publickey;
 if (not(($publickey) = &tcp_slurp($dsock,'LOGINGETPUBLIC')))
 { &infomessbox_w32("Server tcp_slurp problem on login. (1)"); return; }

 my $complexpass = $login_pass.$publickey;
 $complexpass = uc(Digest::SHA::sha1_hex($complexpass)) for (1 .. 7);

 my $pass_answer;
 if (not(($pass_answer) = &tcp_slurp($dsock,'LOGIN~'.$login_name.'/'.$complexpass)))
 { &infomessbox_w32("Server tcp_slurp problem on login. (2)"); return; }

 if ($pass_answer ne 'LOGIN_OK')
 {
  $dsock->close;
  &infomessbox_w32("Authentication problem.\nServer response: ".$pass_answer);
  return;
 }
 return $dsock;
}
#---- tcp_slurp
sub tcp_slurp($$)
{
 my $sock    = $_[0];
 my $tosend  = $_[1];

 my $theselect = IO::Select->new($sock);
 if($theselect->can_write(7))
 {
  #print "can_write\n";
  syswrite($sock,$tosend.$stream_terminator);
  my $sammelrequest;
  while($theselect->can_read(7))
  {
   #print "can_read\n";
   my $data;
   return unless (sysread($sock,$data,1000000));
   $sammelrequest .= $data;
   if ($sammelrequest =~ s/$stream_terminator$//)
   {
    return $sammelrequest;
   }
   #print "must-reread\n";
  }
 }
 close $sock;
 #print "TIMEOUT: end_tcp_slurp\n";
 return;
}
#---- login_slurp
sub login_slurp($$$$$)
{
 my $ip          = $_[0];
 my $port        = $_[1];
 my $login_name  = $_[2];
 my $login_pass  = $_[3];
 my $send_string = $_[4];
 my $ip_port     = join(':',$ip,$port);

 if (not defined $sockhash{$ip_port})
 {
   my $server_answer;
   do
   {
    return unless (($sockhash{$ip_port}) = &getsock_andlogin($ip,$port,$login_name,$login_pass));
    if (not(($server_answer) = &tcp_slurp($sockhash{$ip_port},$send_string)))
    { return if (! &quesmessbox_w32("Server tcp_slurp problem.\nTry to reconnect ?")); }
   }
   while (not defined $server_answer);
   return $server_answer;
 }
 else
 {
  if (my ($return_string) = &tcp_slurp($sockhash{$ip_port},$send_string))
  {
   return $return_string;
  }
  else
  {
   return unless (&quesmessbox_w32("Connection to Server broken.\nError message: $!\n\nTry to reconnect ?"));
   my $server_answer;
   do
   {
    return unless (($sockhash{$ip_port}) = &getsock_andlogin($ip,$port,$login_name,$login_pass));
    if (not(($server_answer) = &tcp_slurp($sockhash{$ip_port},$send_string)))
    { return if (! &quesmessbox_w32("Server tcp_slurp problem.\nTry to reconnect ?")); }
   }
   while (not defined $server_answer);
   return $server_answer;
  }
 }
}

#==== MW CREATE
my $mw = MainWindow->new();
$mw->setIcon(-file => 'images/window_styles_icon.ico');
#my ($exename) = login_slurp($ip,$port,$login_name,$login_pass,"GetWindowFileName~".$hwnd) or die "Cant slurp exename. (".join(",",@ARGV).")";
#my ($classname) = login_slurp($ip,$port,$login_name,$login_pass,"Win32::GUI::GetClassName~".$hwnd) or die "Cant slurp classname. (".join(",",@ARGV).")";
#my ($windowname) = login_slurp($ip,$port,$login_name,$login_pass,"Win32::GuiTest::GetWindowText~".$hwnd) or die "Cant slurp windowname. (".join(",",@ARGV).")";
my ($windata) = login_slurp($ip,$port,$login_name,$login_pass,"MultiCommand|GetWindowFileName~$hwnd|Win32::GUI::GetClassName~$hwnd|Win32::GuiTest::GetWindowText~$hwnd") or die "Cant slurp windowdata. (".join(",",@ARGV).")";
#$mw->title(sprintf('Window Styles for hWnd: %x, %s, %s, %s',$hwnd,$exename,$classname,$windowname));
$mw->title(sprintf('Window Styles for hWnd: %x, %s, %s, %s',$hwnd,split('\|',$windata)));

#==== NOTEBOOK
my $nb = $mw->NoteBook(-focuscolor => 'black')->pack(-expand => 1,-fill => 'both');
#my $page_styles = $nb->add('Styles',-label => 'Window Styles');
#my $page_extstyles = $nb->add('ExtStyles',-label => 'Extended Styles');
#$nb->update;

#==== styles hash
my %extstyles = (
WS_EX_ACCEPTFILES         => [0x00000010,"The window accepts drag-drop files."],
WS_EX_APPWINDOW           => [0x00040000,"Forces a top-level window onto the taskbar when the window is visible. A window does not appear on the taskbar if it is owned by another window (ex: modal-windows) or it has a style like WS_EX_TOOLWINDOW."],
WS_EX_CLIENTEDGE          => [0x00000200,"The window has a border with a sunken edge."],
WS_EX_COMPOSITED          => [0x02000000,"Paints all descendants of a window in bottom-to-top painting order using double-buffering. For more information, see Remarks. This cannot be used if the window has a class style of either CS_OWNDC or CS_CLASSDC. Windows 2000: This style is not supported."],
WS_EX_CONTEXTHELP         => [0x00000400,"The title bar of the window includes a question mark. When the user clicks the question mark, the cursor changes to a question mark with a pointer. If the user then clicks a child window, the child receives a WM_HELP message. The child window should pass the message to the parent window procedure, which should call the WinHelp function using the HELP_WM_HELP command. The Help application displays a pop-up window that typically contains help for the child window. WS_EX_CONTEXTHELP cannot be used with the WS_MAXIMIZEBOX or WS_MINIMIZEBOX styles."],
WS_EX_CONTROLPARENT       => [0x00010000,"The window itself contains child windows that should take part in dialog box navigation. If this style is specified, the dialog manager recurses into children of this window when performing navigation operations such as handling the TAB key, an arrow key, or a keyboard mnemonic."],
WS_EX_DLGMODALFRAME       => [0x00000001,"The window has a double border; the window can, optionally, be created with a title bar by specifying the WS_CAPTION style in the dwStyle parameter."],
WS_EX_LAYERED             => [0x00080000,"The window is a layered window. This style cannot be used if the window has a class style of either CS_OWNDC or CS_CLASSDC. Windows 8: The WS_EX_LAYERED style is supported for top-level windows and child windows. Previous Windows versions support WS_EX_LAYERED only for top-level windows."],
WS_EX_LAYOUTRTL           => [0x00400000,"If the shell language is Hebrew, Arabic, or another language that supports reading order alignment, the horizontal origin of the window is on the right edge. Increasing horizontal values advance to the left."],
WS_EX_LEFT                => [0x00000000,"The window has generic left-aligned properties. This is the default."],
WS_EX_LEFTSCROLLBAR       => [0x00004000,"If the shell language is Hebrew, Arabic, or another language that supports reading order alignment, the vertical scroll bar (if present) is to the left of the client area. For other languages, the style is ignored."],
WS_EX_LTRREADING          => [0x00000000,"The window text is displayed using left-to-right reading-order properties. This is the default."],
WS_EX_MDICHILD            => [0x00000040,"The window is a MDI child window."],
WS_EX_NOACTIVATE          => [0x08000000,"A top-level window created with this style does not become the foreground window when the user clicks it. The system does not bring this window to the foreground when the user minimizes or closes the foreground window. To activate the window, use the SetActiveWindow or SetForegroundWindow function. The window does not appear on the taskbar by default. To force the window to appear on the taskbar, use the WS_EX_APPWINDOW style."],
WS_EX_NOINHERITLAYOUT     => [0x00100000,"The window does not pass its window layout to its child windows."],
WS_EX_NOPARENTNOTIFY      => [0x00000004,"The child window created with this style does not send the WM_PARENTNOTIFY message to its parent window when it is created or destroyed."],
WS_EX_NOREDIRECTIONBITMAP => [0x00200000,"The window does not render to a redirection surface. This is for windows that do not have visible content or that use mechanisms other than surfaces to provide their visual."],
WS_EX_OVERLAPPEDWINDOW    => [0x00000300,"(WS_EX_WINDOWEDGE | WS_EX_CLIENTEDGE)\nThe window is an overlapped window."],
WS_EX_PALETTEWINDOW       => [0x00000188,"(WS_EX_WINDOWEDGE | WS_EX_TOOLWINDOW | WS_EX_TOPMOST)\nThe window is palette window, which is a modeless dialog box that presents an array of commands."],
WS_EX_RIGHT               => [0x00001000,"The window has generic \"right-aligned\" properties. This depends on the window class. This style has an effect only if the shell language is Hebrew, Arabic, or another language that supports reading-order alignment; otherwise, the style is ignored. Using the WS_EX_RIGHT style for static or edit controls has the same effect as using the SS_RIGHT or ES_RIGHT style, respectively. Using this style with button controls has the same effect as using BS_RIGHT and BS_RIGHTBUTTON styles."],
WS_EX_RIGHTSCROLLBAR      => [0x00000000,"The vertical scroll bar (if present) is to the right of the client area. This is the default."],
WS_EX_RTLREADING          => [0x00002000,"If the shell language is Hebrew, Arabic, or another language that supports reading-order alignment, the window text is displayed using right-to-left reading-order properties. For other languages, the style is ignored."],
WS_EX_STATICEDGE          => [0x00020000,"The window has a three-dimensional border style intended to be used for items that do not accept user input."],
WS_EX_TOOLWINDOW          => [0x00000080,"The window is intended to be used as a floating toolbar. A tool window has a title bar that is shorter than a normal title bar, and the window title is drawn using a smaller font. A tool window does not appear in the taskbar or in the dialog that appears when the user presses ALT+TAB. If a tool window has a system menu, its icon is not displayed on the title bar. However, you can display the system menu by right-clicking or by typing ALT+SPACE."],
WS_EX_TOPMOST             => [0x00000008,"The window should be placed above all non-topmost windows and should stay above them, even when the window is deactivated. To add or remove this style, use the SetWindowPos function."],
WS_EX_TRANSPARENT         => [0x00000020,"The window should not be painted until siblings beneath the window (that were created by the same thread) have been painted. The window appears transparent because the bits of underlying sibling windows have already been painted. To achieve transparency without these restrictions, use the SetWindowRgn function."],
WS_EX_WINDOWEDGE          => [0x00000100,"The window has a border with a raised edge."]);
my %styles = (
WS_BORDER           => [0x00800000,"The window has a thin-line border."],
WS_CAPTION          => [0x00C00000,"The window has a title bar (includes the WS_BORDER style)."],
WS_CHILD            => [0x40000000,"The window is a child window. A window with this style cannot have a menu bar. This style cannot be used with the WS_POPUP style."],
WS_CHILDWINDOW      => [0x40000000,"Same as the WS_CHILD style."],
WS_CLIPCHILDREN     => [0x02000000,"Excludes the area occupied by child windows when drawing occurs within the parent window. This style is used when creating the parent window."],
WS_CLIPSIBLINGS     => [0x04000000,"Clips child windows relative to each other; that is, when a particular child window receives a WM_PAINT message, the WS_CLIPSIBLINGS style clips all other overlapping child windows out of the region of the child window to be updated. If WS_CLIPSIBLINGS is not specified and child windows overlap, it is possible, when drawing within the client area of a child window, to draw within the client area of a neighboring child window."],
WS_DISABLED         => [0x08000000,"The window is initially disabled. A disabled window cannot receive input from the user. To change this after a window has been created, use the EnableWindow function."],
WS_DLGFRAME         => [0x00400000,"The window has a border of a style typically used with dialog boxes. A window with this style cannot have a title bar."],
WS_GROUP            => [0x00020000,"The window is the first control of a group of controls. The group consists of this first control and all controls defined after it, up to the next control with the WS_GROUP style. The first control in each group usually has the WS_TABSTOP style so that the user can move from group to group. The user can subsequently change the keyboard focus from one control in the group to the next control in the group by using the direction keys. You can turn this style on and off to change dialog box navigation. To change this style after a window has been created, use the SetWindowLong function."],
WS_HSCROLL          => [0x00100000,"The window has a horizontal scroll bar."],
WS_ICONIC           => [0x20000000,"The window is initially minimized. Same as the WS_MINIMIZE style."],
WS_MAXIMIZE         => [0x01000000,"The window is initially maximized."],
WS_MAXIMIZEBOX      => [0x00010000,"The window has a maximize button. Cannot be combined with the WS_EX_CONTEXTHELP style. The WS_SYSMENU style must also be specified."],
WS_MINIMIZE         => [0x20000000,"The window is initially minimized. Same as the WS_ICONIC style."],
WS_MINIMIZEBOX      => [0x00020000,"The window has a minimize button. Cannot be combined with the WS_EX_CONTEXTHELP style. The WS_SYSMENU style must also be specified."],
WS_OVERLAPPED       => [0x00000000,"The window is an overlapped window. An overlapped window has a title bar and a border. Same as the WS_TILED style."],
WS_OVERLAPPEDWINDOW => [0x00CF0000,"(WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME | WS_MINIMIZEBOX | WS_MAXIMIZEBOX)\nThe window is an overlapped window. Same as the WS_TILEDWINDOW style."],
WS_POPUP            => [0x80000000,"The windows is a pop-up window. This style cannot be used with the WS_CHILD style."],
WS_POPUPWINDOW      => [0x80880000,"(WS_POPUP | WS_BORDER | WS_SYSMENU)\nThe window is a pop-up window. The WS_CAPTION and WS_POPUPWINDOW styles must be combined to make the window menu visible."],
WS_SIZEBOX          => [0x00040000,"The window has a sizing border. Same as the WS_THICKFRAME style."],
WS_SYSMENU          => [0x00080000,"The window has a window menu on its title bar. The WS_CAPTION style must also be specified."],
WS_TABSTOP          => [0x00010000,"The window is a control that can receive the keyboard focus when the user presses the TAB key. Pressing the TAB key changes the keyboard focus to the next control with the WS_TABSTOP style. You can turn this style on and off to change dialog box navigation. To change this style after a window has been created, use the SetWindowLong function. For user-created windows and modeless dialogs to work with tab stops, alter the message loop to call the IsDialogMessage function."],
WS_THICKFRAME       => [0x00040000,"The window has a sizing border. Same as the WS_SIZEBOX style."],
WS_TILED            => [0x00000000,"The window is an overlapped window. An overlapped window has a title bar and a border. Same as the WS_OVERLAPPED style."],
WS_TILEDWINDOW      => [0x00CF0000,"(WS_OVERLAPPED | WS_CAPTION | WS_SYSMENU | WS_THICKFRAME | WS_MINIMIZEBOX | WS_MAXIMIZEBOX)\nThe window is an overlapped window. Same as the WS_OVERLAPPEDWINDOW style."],
WS_VISIBLE          => [0x10000000,"The window is initially visible. This style can be turned on and off by using the ShowWindow or SetWindowPos function."],
WS_VSCROLL          => [0x00200000,"The window has a vertical scroll bar."]);

#==== zum setzen des stylestatus der checkboxen
sub set_checkbox_state()
{
 if (my ($styledata) = login_slurp($ip,$port,$login_name,$login_pass,"MultiCommand|Win32::GUI::GetWindowLong~$hwnd~-16|Win32::GUI::GetWindowLong~$hwnd~-20"))
 {
  my ($style,$extstyle) = split('\|',$styledata);
  printf("style: %008x\n",$style);
  foreach my $styleid (keys %styles)
  {
   $styles{$styleid}[2] = ($styles{$styleid}[0] == 0 || ($style & $styles{$styleid}[0]) == $styles{$styleid}[0])?1:0;
  }
  foreach my $styleid (keys %extstyles)
  {
   $extstyles{$styleid}[2] = ($extstyles{$styleid}[0] == 0 || ($extstyle & $extstyles{$styleid}[0]) == $extstyles{$styleid}[0])?1:0;
  }
 }
}

#==== zum setzen des styles
sub set_style()
{
 my $styles = ($_[1])?\%extstyles:\%styles;
 return set_checkbox_state &infomessbox_w32("You cannot remove the standard style id zero.") if ($styles->{$_[0]}[0] == 0);
 if (my ($style) = login_slurp($ip,$port,$login_name,$login_pass,"Win32::GUI::GetWindowLong~$hwnd~".(($_[1])?-20:-16)))
 {
  my ($ret) = login_slurp($ip,$port,$login_name,$login_pass,"Win32::GUI::SetWindowLong~$hwnd~".(($_[1])?-20:-16)."~".(($styles->{$_[0]}[2])?$style|$styles->{$_[0]}[0]:$style&~$styles->{$_[0]}[0])) or return;
  my ($ret) = login_slurp($ip,$port,$login_name,$login_pass,"Win32::GUI::SetWindowPos~$hwnd~0~0~0~0~0~".0x227) or return;
  set_checkbox_state();
 }
}

#==== create window styles tk frame
sub write_tab($)
{
 my $styles = ($_[0])?\%extstyles:\%styles;
 my $page_styles = $nb->add('Styles'.$_[0],-label => ($_[0])?'Extended Window Styles':'Window Styles');
 my $styles_frame = $page_styles->Scrolled('Frame',-sticky => "w",-scrollbars => "e",-height => 450)->pack(-expand => 1,-fill => 'both');
 $styles_frame->bind("<MouseWheel>",[sub{$styles_frame->yview('scroll',($_[1]==120)?-5:5,'units');},Ev('D')]);
 $styles_frame->bind("<Enter>",sub{$_[0]->focus;});
 my $row = 0;
 $styles_frame->Frame(-height => 3,-bg => "grey70",-relief => 'sunken')->grid(-column => 0,-columnspan => 3,-row => $row++,-sticky => 'we');
 foreach my $styleid (sort keys %$styles)
 {
  $styles_frame->Checkbutton(-command => [\&set_style,$styleid,$_[0]],-pady => 5,-variable=>\$styles->{$styleid}[2])->grid(-column => 0,-row => $row,-sticky => '');
  my $puffer_frame = $styles_frame->Frame(-pady => 5)->grid(-column => 1,-row => $row,-sticky => 'w');
  $puffer_frame->Label(-text => $styleid,-font => [-size => 10,-weight => "bold"],-pady => 0,-justify => "left")->grid(-column => 0,-row => 0,-sticky => 'w');
  $puffer_frame->Label(-text => sprintf("0x%08XL",$styles->{$styleid}[0]),-pady => 0,-justify => "left",-wraplength => 400)->grid(-column => 0,-row => 1,-sticky => 'w');
  $styles_frame->Label(-text => $styles->{$styleid}[1],-pady => 5,-justify => "left",-wraplength => 400)->grid(-column => 2,-row => $row++,-sticky => 'w');
  $styles_frame->Frame(-height => 3,-bg => "grey70",-relief => 'sunken')->grid(-column => 0,-columnspan => 3,-row => $row++,-sticky => 'we');
 }
}
write_tab(0);
write_tab(1);

#==== check boxes
set_checkbox_state();

###################################################################################
######################################################################## Mainloop #
###################################################################################
MainLoop();
